#! /usr/bin/env ruby

require 'bindata'
require 'ostruct'
require 'openssl'
require 'bcrypt_pbkdf'

CIPHERS = {
  'aes256-ctr' => OpenStruct.new(block_size: 16, key_len: 32, iv_len: 16, name: 'AES-256-CTR')
}

CURVES = {
  'nistp256' => 'prime256v1'
}

def decode_binary(data)
  return nil if data.empty?
  data.each_char.inject(0) { |a, b| (a << 8) | b.ord }
end

# A byte string prefixed by 32 bit length
class String32 < BinData::Primitive
  endian :big
  uint32 :len
  string :data, read_length: :len

  def get
    data
  end
end

# Bignum32, a variant of String32 that displays as a bignum.
class Bn32 < String32
  def get
    decode_binary(data)
  end
end

class SSHKeyHeader < BinData::Record
  endian :big

  stringz :format # "openssh-key-v1"
  string32 :cipher # "none" or "aes256-ctr" in recent versions of openssh
  string32 :kdf # "none" or "bcrypt"
end

class ECPoint < BinData::Primitive
  mandatory_parameter :curve
  string32 :q

  def group
    curve = eval_parameter(:curve).to_s
    ec = OpenSSL::PKey::EC.new(CURVES[curve])
    ec.group
  end

  def get
    OpenSSL::PKey::EC::Point.new(group, q.to_s)
  end

  # Workaround for bindata, which compares it to nil to present sensible_default instead
  # and OpenSSL classes actively resent being compared to nil
  def _value
    get
  end
end

class ECDSAPubkey < BinData::Record
  endian :big
  string32 :key_type
  string32 :curve
  ec_point :q, curve: :curve, assert: -> { q.on_curve? }
  string32 :application, assert: 'ssh:'
end

class ECDSAPrivkey < BinData::Record
  endian :big
  uint32 :check_a
  uint32 :check_b
  string32 :key_type
  string32 :curve
  ec_point :q, curve: :curve, assert: -> { q.on_curve? }
  string32 :application, assert: 'ssh:'
  int8 :flags
  string32 :key_handle
  string32 :reserved
end

class PlainSSHKey < BinData::Record
  endian :big

  bn32 # ignore kdf_data
  uint32 :num_keys, assert: 1
  uint32 :pubkeys_len
  array :pubkeys, type: :ecdsa_pubkey, initial_length: :num_keys
  uint32 :privkey_len
  ecdsa_privkey :privkey
  string32 :comment
  # And padding
end

class BcryptData < BinData::Record
  endian :big
  uint32 # bcdata_len, not useful
  string32 :salt
  uint32 :rounds
end

class EncryptedSSHKey < BinData::Record
  mandatory_parameter :cipher
  endian :big

  bcrypt_data :kdf_data
  uint32 :num_keys, assert: 1
  uint32 :pubkeys_len # length of following pubkey block
  array :pubkeys, type: :ecdsa_pubkey, initial_length: :num_keys

  uint32 :enc_len
  virtual :assert_blocklen, assert: -> { enc_len % CIPHERS[cipher.to_s].block_size == 0 }
  string :enc_privkey, read_length: :enc_len # length of following privkey block

  def salt
    kdf_data.salt.to_s
  end

  def passphrase
    ARGV[0] or raise 'Must provide passphrase on commandline'
  end

  def key_and_iv
    key_len = cipher_params.key_len
    key_with_iv = BCryptPbkdf.key(passphrase, salt,
                                  key_len + cipher_params.iv_len, kdf_data.rounds)
    shield_key = key_with_iv[0...key_len]
    iv = key_with_iv[key_len..]

    [shield_key, iv]
  end

  def cipher_params
    CIPHERS[get_parameter(:cipher)]
  end

  def privkey
    decoder = OpenSSL::Cipher.new(cipher_params.name).decrypt
    decoder.key, decoder.iv = key_and_iv
    data = decoder.update(enc_privkey.to_s)
    ECDSAPrivkey.read(data)
  end
end

header = SSHKeyHeader.read(STDIN)
cipher = header.cipher.to_s
key = case cipher
when 'none'
  PlainSSHKey.read(STDIN)
else
  # Not passing kdf, as we only support bcrypt anyway
  EncryptedSSHKey.read(STDIN, cipher: cipher)
end

# Validating the file
pub = OpenStruct.new(key[:pubkeys].first)
pp key
pp key.privkey

fail "Public and private key must have identical qvalue" unless pub.q == key.privkey.q
fail "key-handle must not be empty" unless key.privkey.key_handle
